-- Importer for Wii U BFRES files seen in many Nintendo games. 
-- Written by ItsEasyActually and Random Talking Bush (with some help from Ploaj for 10-bit normal values, and Sylk for optimizations).
-- Credit to the tockdom wiki community for the format documention
-- Special thanks to TGE for help early on with new scripting methods and getting ItsEasyActually started

-- Changelog:
-- May 12th, 2017:
-- Script has been optimized so that models will import significantly faster, thanks to Sylk!
-- Changed "mat = standardMaterial()" to "local PolyMat = standardMaterial()" to fix potential problems, as suggested by Sylk.
-- Shaved off about a thousand lines of redundant coding since the way vertex buffers are read has now been rewritten.
-- Added an option to print debug information to the Listener, so the separate "_Debug" script is no longer necessary.

-- April 7th, 2017:
-- Fixed UV mapping for models with type 516 (e.g. Dry Bowser's eyes in M&S2014.

-- April 2nd, 2017:
-- Fixed UV mapping for models with type 519 (e.g. Riju's model in BotW). Again.

-- March 11th, 2017:
-- Fixed importing for Wind Waker HD models, which wasn't set up properly in the script in the first place. Whoops!

-- March 6th, 2017:
-- Fixed rigging being utterly broken in 3DS Max 2012 and above (thanks to sunnydavis for part of the code used). Gee, *thanks*, Autodesk...
-- Also fixed rigging for models with more than 4 bone influences (Zelda's model in BotW, for example).
-- Added support for reading normals, thanks to Ploaj for helping with the 10-bit nonsense!
-- Added type 522 for colours (signed bytes, e.g. Fire Peach's dress model in SM3DW).

-- October 15th, 2016:
-- Added a UI to choose which BFMDL to import in a BFRES, and various little options.
-- LOD (Level of Detail) models can now be imported.
-- Fixed Version 0 BFRES bone structures so they're no longer broken (NSMBU, NintendoLand)
-- Fixed Version 3 BFRES bone structures so they actually load now (Wii Sports Resort)
-- Fixed MatrFlag-related importing so now it won't take AGES to load maps!
-- Fixed texture importing for certain Paper Mario: Color Splash models.
-- Fixed vertex colours which use full floats instead of half floats (LoZWWHD), thanks DSX8!
-- Added UV mapping support for type 516 (bytes), might be incorrect.
-- Not a fix, but UV mapping type 2056 is currently incorrect and will need to be looked into later.
-- Likewise, normals will be looked into for the next update.


rollout BFRESImporter "Wii U BFRES model importer" width:390 height:225
(

	label lblDisclaimer "This script was written by ItsEasyActually and Random Talking Bush, with some assistance from both Ploaj and Sylk. If you use it, consider giving us all thanks for this. If something doesn't work right, please contact RTB on The VG Resource (Random Talking Bush), Twitter, Tumblr or Steam (RandomTBush) so that any problems can be fixed." pos:[8,8] width:375 height:75

	button btnLoad "Load BFRES" pos:[8,85] width:90 height:20
	button btnImport "Import BFMDL" pos:[108,85] width:90 height:20
	dropdownlist lstFMDL pos:[208,85] width:175 height:20 items:#() selection:0

	label lblOptions "Options:" pos:[8,110] width:375 height:15
	label lblTexFormat "Texture format:" pos:[8,125] width:375 height:15
	radiobuttons tglTexFormat labels:#("DDS","PNG") pos:[90,125]
	
	label lblLODs "Import LODs:" pos:[8,140] width:375 height:15
	radiobuttons tglLODs labels:#("No","Yes") pos:[90,140]

	label lblUVLayers "UV layering:" pos:[8,155] width:375 height:15
	radiobuttons tglUVLayers labels:#("Split","Merge","No") pos:[90,155]

	label lblRigModel "Import rigging:" pos:[8,170] width:375 height:15
	radiobuttons tglRigModel labels:#("Yes","No") pos:[90,170]
	checkbox tglDebug "Print debug information to Listener?" pos:[8,185] width:375 height:15

	label lblUpdate "(Updated 05/13/2017) " pos:[8,210] width:375 height:15


fn ReadOffset fstream = ((ftell fstream) + (bit.swapBytes (bit.swapBytes (readlong fstream #unsigned) 1 4) 2 3))
fn readBEshort fstream = (bit.swapBytes (readshort fstream #unsigned) 1 2)
fn readBElong fstream = (bit.swapBytes (bit.swapBytes (readlong fstream #unsigned) 1 4) 2 3)
fn readBEfloat fstream = (bit.intAsFloat (bit.swapBytes (bit.swapBytes (readlong fstream #unsigned) 1 4) 2 3))

fn ReadBEHalfFloat Fstream = (
	local BH  = readByte Fstream #unsigned
	local BL  = readByte Fstream #unsigned
	local N = BH*256 + BL
	local S = floor((mod N 65536) / 32768)
	local Ef = floor((mod N 32768) / 1024)
	local M = mod N 1024
	local ret=case of(
		((Ef==0)AND(M==0)): ( (-1.0)^S * 0.0 )
		((Ef==0)AND(M!=0)): ( (-1.0)^S * 2.0^-14 * (M / 2.0^10) )
		((Ef>0)AND(Ef<31)): ( (-1.0)^S * 2.0^(Ef-15) * (1 + M/2.0^10) )
		((Ef==31)AND(M==0)): ( (-1.0)^S * 1/0.0 )
		((Ef==31)AND(M!=0)): 0 --hack-- should be #inf
	)
)

fn sign10 x = (
	if (bit.shift x -9) == 1 do(
		x = bit.not x
		x = bit.and x 0x3FF
		x += 1
		x *= -1
	)
	x
) --Thanks, Ploaj!

fn ReadFixedString fstream fixedLen = (
    local str = ""
    for i = 1 to fixedLen do(str += bit.intAsChar (ReadByte fstream #unsigned))
    str
)

fn vRange Lo Hi val=(
    val=case of(
        (val>Hi):Hi
        (val<Lo):Lo
        default:val
    )
)

	local FMDLOffset_array = #()
	local f = #()
	local p = #()


on btnLoad pressed do(
clearlistener()
fname = GetOpenFileName caption:"Open BFRES" types:"BFRES File(*.bfres)|*.bfres"
st = timestamp() --get start time in milliseconds

if fname != undefined do(
FMDLName_array = #()
FMDLOffset_array = #()

p = getFilenamePath fname
f = fopen fname "rb"   --open file in read only format

	struct fmdlh(
		fmdl,				-- Magic
		fnameOff,		-- Filename Offset
		eofString,		-- End of File String Offset
		fsklOff,			-- FSKL Offset
		fvtxArrOff,		-- FVTX Array Offset
		fshpIndx,		-- FSHP Index Offset
		fmatIndx,		-- FMAT Index Offset
		paramOff,		-- Parameter Offset
		fvtxCount,		-- FVTX Count (int16)
		fshpCount,		-- FSHP Count (int16)
		fmatCount,		-- FMAT Count (int16
		paramCount	-- Parameter Count (int16)
	)

fseek f 5#seek_set
verNum = ReadByte f
print ("FMDL Version: " + verNum as string)
fseek f 26#seek_cur
FileOffset = ReadOffset f
fseek f FileOffset#seek_set
BlockSize = ReadBELong f
FMDLTotal = ReadBELong f
fseek f 0x10#seek_cur

For mdl = 1 to FMDLTotal Do (
	fseek f 12#seek_cur
	FMDLOffset = ReadOffset f
	append FMDLOffset_array FMDLOffset
	NextFMDL = ftell f

	fseek f FMDLOffset#seek_set

	-- F_Model Header --
	fmdl_info=(fmdlh \
		fmdl: (ReadFixedString f 4) \
		fnameOff: (ReadOffset f) \
		eofString: (ReadOffset f) \
		fsklOff: (ReadOffset f) \
		fvtxArrOff: (ReadOffset f) \
		fshpIndx: (ReadOffset f) \
		fmatIndx: (ReadOffset f) \
		paramOff: (ReadOffset f) \
		fvtxCount: (ReadBEShort f) \
		fshpCount: (ReadBEShort f) \
		fmatCount: (ReadBEShort f) \
		paramCount: (ReadBEShort f) \
	)

	fseek f fmdl_info.fnameOff#seek_set
	FMDLName = ReadString f
	append FMDLName_array FMDLName
	fseek f NextFMDL#seek_set
)

lstFMDL.items = FMDLName_array
lstFMDL.selection = 1

)
)

on btnImport pressed do(
if lstFMDL.selection != 0 do(
clearlistener()
st = timestamp() --get start time in milliseconds
disableSceneRedraw()

	struct fmdlh(
		fmdl,				-- Magic
		fnameOff,		-- Filename Offset
		eofString,		-- End of File String Offset
		fsklOff,			-- FSKL Offset
		fvtxArrOff,		-- FVTX Array Offset
		fshpIndx,		-- FSHP Index Offset
		fmatIndx,		-- FMAT Index Offset
		paramOff,		-- Parameter Offset
		fvtxCount,		-- FVTX Count (int16)
		fshpCount,		-- FSHP Count (int16)
		fmatCount,		-- FMAT Count (int16
		paramCount	-- Parameter Count (int16)
	)

	struct fvtxh(
		fvtx,				-- Magic
		attCount,		-- Attribute Count (int8)
		buffCount,		-- Buffer Count (int8)
		sectIndx,		-- Section Index (int16)
		vertCount,		-- Vertex Count
		u1,				-- Unknown (int8 or int16)
		u2,				-- Unknown (int16 or three bytes)
		attArrOff,		-- Attribute Array Offset
		attIndxOff,		-- Attribute Index Offset
		buffArrOff,		-- Buffer Array Offset
		padding			-- Padding (int32)
	)

	struct fmath(
		fmat,						-- Magic
		matOff,					-- Material Name Offset
		u1,						-- Unknown (int32)
		sectIndx,				-- Section Index (int16)
		rendParamCount,		-- Render Info Parameter Count (int16)
		texSelCount,			-- Texture Selector Count (int8)
		texAttSelCount,		-- Texture Attribute Selector Count (int8)
		matParamCount,		-- Material Parameter Count (int16)
		matParamSize,			-- Material Parameter Data Size
		u2,						-- Unknown (int32)
		rendParamIndx,		-- Render Info Parameter Index Group Offset
		unkMatOff,				-- Unknown Material Offset
		shadeOff,				-- Shader Control Offset
		texSelOff,				-- Texture Selector Offset
		texAttSelOff,			-- Texture Attribute Selector Offset
		texAttIndxOff,			-- Texture Attribute Selector Index Offset
		matParamArrOff,		-- Material Parameter Array Offset
		matParamIndxOff,	-- Material Parameter Index Offset
		matParamOff,			-- Material Parameter Data Offset
		shadParamIndxOff	-- Shadow Parameter Index Offset
	)

	struct fsklh(
		fskl,					-- Magic
		u1,					-- Unknown (int16)
		fsklType,			-- Type of Bone Data (uint16)
		boneArrCount,		-- Bone Array Count (int16)
		invIndxArrCount,	-- Inverse Index Array Count (int16)
		exIndxCount,		-- Extra Index Count (int16)
		u3,					-- Unknown (int16)
		boneIndxOff,		-- Bone Index Group Offset
		boneArrOff,			-- Bone Array Offset
		invIndxArrOff,		-- Inverse Index Array Offset
		invMatrArrOff,		-- Inverse Matric Array Offset
		padding				-- Padding (int32)
	)

	struct fshph(
		fshp,					-- Magic
		polyNameOff,		-- Polygon Name Offset
		u1,					-- Unknown (int32)
		fvtxIndx,			-- FVTX Index (int16)
		fmatIndx,			-- FMAT Index (int16)
		fsklIndx,				-- FSKL Index (int16)
		sectIndx,			-- Section Index (int16)
		fsklIndxArrCount,	-- FSKL Index Array Count (int16)
		matrFlag,			-- Matrix Use Flag (int8)
		lodMdlCount,		-- LoD Model Count (int8)
		visGrpCount,		-- Visibility Group Tree Node Count
		u3,					-- Unknown (float)
		fvtxOff,				-- FVTX Offset
		lodMdlOff,			-- LoD Model Offset
		fsklIndxArrOff,		-- FSKL Index Array Offset
		u4,					-- Unknown (int32)
		visGrpNodeOff,	-- Vibility Group Tree Nodes Offset
		visGrpRangeOff,	-- Visibility Group Tree Ranges Offset
		visGrpIndxOff,		-- Visibility Group Tree Indices Offset
		u5						-- Unknown (int32)
	)	

	struct attdata(
		attName,	-- Attribute Type
		buffIndx,	-- Buffer Index (u8)
		buffOff,		-- Buffer Offset (u16)
		vertType	-- Format
	)

	struct buffData(
		buffSize,	-- Size of Buffer (u32)
		strideSize,	-- Size of Vertex Stride (u16)
		dataOffset	-- Offset to Buffer (u32)
	)

	struct bdata(
		bNameOff,		-- Bone Name Offset
		bIndx,			-- Bone Index (int16)
		parIndx1,		-- Parent Index 1 (int16)
		parIndx2,		-- Parent Index 2 (int16)
		parIndx3,		-- Parent Index 3 (int16)
		parIndx4,		-- Parent Index 4 (int16)
		u1,				-- Unknown (int16)
		bFlags,			-- Bone Flags (int16)
		u2,				-- Unknown (int16)
		scaleX,			-- Scale X
		scaleY,			-- Scale Y
		scaleZ,			-- Scale Z
		rotX,				-- Rotation X
		rotY,				-- Rotation Y
		rotZ,				-- Rotation Z
		rotW,				-- Rotation W
		posX,				-- Position X
		posY,				-- Position Y
		posZ,				-- Position Z
		padding,			-- Padding
		m11,				-- Matrix 
		m21,				-- Matrix 
		m31,				-- Matrix 
		m41,				-- Matrix 
		m12,				-- Matrix 
		m22,				-- Matrix 
		m32,				-- Matrix 
		m42,				-- Matrix 
		m13,				-- Matrix 
		m23,				-- Matrix 
		m33,				-- Matrix 
		m43				-- Matrix 
	)

	struct lodmdl(
		u1,				-- Unknown (int32)
		faceType,		-- Face Data Type (int32)
		dCount,			-- Visibility Group Draw Count
		visGrpCount,	-- Visibility Group Count (int16)
		u3,				-- Unknown (int16)
		visGrpOff,		-- Visibility Group Offset
		indxBuffOff,		-- Index Buffer Offset
		elmSkip			-- Element Skip (int32)
	)

	struct Bone_Info_Struct
	(
		bone1, bone2, bone3, bone4
	)

	struct Weight_Info_Struct
	(
		weight1, weight2, weight3, weight4
	)

	struct weight_data
	(
		boneids, weights
	)

	struct invMatrix
	(
		invIndx, invMatr
	)
	

	local TexFormat = tglTexFormat.state
	local LODs = tglLODs.state
	local UVLayers = tglUVLayers.state
	local RigModel = tglRigModel.state
	local PrintDebug = tglDebug.checked

	fseek f 5#seek_set
	verNum = ReadByte f

	fseek f FMDLOffset_array[lstFMDL.selection]#seek_set

	FMDLArr = #()
	FVTXArr = #()
	FSKLArr = #()
	FMATArr = #()
	FMATNameArr = #()
	FSHPArr = #()
	VTXAttr = #()
	
	BoneArray = #()
	BoneFixArray = #()
	invIndxArr = #()
	invMatrArr = #()
	Node_Array = #()

	-- F_Model Header --
	fmdl_info=(fmdlh \
		fmdl: (ReadFixedString f 4) \
		fnameOff: (ReadOffset f) \
		eofString: (ReadOffset f) \
		fsklOff: (ReadOffset f) \
		fvtxArrOff: (ReadOffset f) \
		fshpIndx: (ReadOffset f) \
		fmatIndx: (ReadOffset f) \
		paramOff: (ReadOffset f) \
		fvtxCount: (ReadBEShort f) \
		fshpCount: (ReadBEShort f) \
		fmatCount: (ReadBEShort f) \
		paramCount: (ReadBEShort f) \
	)
	append FMDLArr(fmdl_info)
	if PrintDebug == true do(print (fmdl_info))

	-- F_Vertex Header --
	fseek f fmdl_info.fvtxArrOff#seek_set
	For vtx = 1 to fmdl_info.fvtxCount Do (
		fvtx_info=(fvtxh \
			fvtx: (ReadFixedString f 4) \
			attCount: (ReadByte f) \
			buffCount: (ReadByte f) \
			sectIndx: (ReadBEshort f) \
			vertCount: (ReadBELong f) \
			u1: (ReadBEShort f) \
			u2: (ReadBEShort f) \
			attArrOff: (ReadOffset f) \
			attIndxOff: (ReadOffset f) \
			buffArrOff: (ReadOffset f) \
			padding: (ReadLong f) \
		)
		append FVTXArr(fvtx_info)
		if PrintDebug == true do(print (fvtx_info))
	)

	-- F_Material Header --
	fseek f fmdl_info.fmatIndx#seek_set
	fseek f 24#seek_cur
	For mat = 1 to fmdl_info.fmatCount Do (
		fseek f 8#seek_cur
		FMATNameOffset = ReadOffset f
		Rtn = ftell f
		fseek f FMATNameOffset#seek_set
		
		FMATName = readstring f
		append FMATNameArr FMATName
		fseek f Rtn#seek_set
		
		FMATOffset = ReadOffset f
		Rtn = ftell f
		
		fseek f FMATOffset#seek_set
		fmat_info=(fmath \
			fmat: (ReadFixedString f 4) \
			matOff: (ReadOffset f) \
			u1: (ReadBELong f) \
			sectIndx: (ReadBEShort f) \
			rendParamCount: (ReadBEshort f) \
			texSelCount: (ReadByte f) \
			texAttSelCount: (ReadByte f) \
			matParamCount: (ReadBEShort f) \
			matParamSize: (ReadBELong f) \
			u2: (ReadBELong f) \
			rendParamIndx: (ReadOffset f) \
			unkMatOff: (ReadOffset f) \
			shadeOff: (ReadOffset f) \
			texSelOff: (ReadOffset f) \
			texAttSelOff: (ReadOffset f) \
			texAttIndxOff: (ReadOffset f) \
			matParamArrOff: (ReadOffset f) \
			matParamIndxOff: (ReadOffset f) \
			matParamOff: (ReadOffset f) \
			shadParamIndxOff: (ReadOffset f) \
		)
		
		fseek f Rtn#seek_set
		append FMATArr(fmat_info)
		if PrintDebug == true do(print fmat_info)
	)

	-- F_Skeleton Header --
	fseek f fmdl_info.fsklOff#seek_set
	fskl_info=(fsklh \
		fskl: (ReadFixedString f 4) \
		u1: (ReadBEShort f) \
		fsklType: (ReadBEShort f) \
		boneArrCount: (ReadBEShort f) \
		invIndxArrCount: (ReadBEShort f) \
		exIndxCount: (ReadBEShort f) \
		u3: (ReadBEShort f) \
		boneIndxOff: (ReadOffset f) \
		boneArrOff: (ReadOffset f) \
		invIndxArrOff: (ReadOffset f) \
		invMatrArrOff: (ReadOffset f) \
		padding: (ReadLong f) \
	)
	append FSKLArr(fskl_info)
	if PrintDebug == true do(print (fskl_info))

	-- Node Setup --
	fseek f fskl_info.invIndxArrOff #seek_set
	for nodes = 1 to (fskl_info.invIndxArrCount + fskl_info.exIndxCount) do(
		NodeID = readBEshort f + 1
		append Node_Array NodeID
	)

	-- F_Shape Header --
	fseek f fmdl_info.fshpIndx#seek_set
	fseek f 24#seek_cur
	For shp = 1 to fmdl_info.fshpCount Do (
		fseek f 12#seek_cur
		FSHPOffset = ReadOffset f
		Rtn = ftell f
		
		fseek f FSHPOffset#seek_set
		fshp_info=(fshph \
			fshp: (ReadFixedString f 4) \
			polyNameOff: (ReadOffset f) \
			u1: (ReadBELong f) \
			fvtxIndx: (ReadBEShort f +1) \
			fmatIndx: (ReadBEShort f +1) \
			fsklIndx: (ReadBEShort f +1) \
			sectIndx: (ReadBEShort f +1) \
			fsklIndxArrCount: (ReadBEShort f) \
			matrFlag: (ReadByte f) \
			lodMdlCount: (ReadByte f) \
			visGrpCount: (ReadBELong f) \
			u3: (ReadBEFloat f) \
			fvtxOff: (ReadOffset f) \
			lodMdlOff: (ReadOffset f) \
			fsklIndxArrOff: (ReadOffset f) \
			u4: (ReadBELong f) \
			visGrpNodeOff: (ReadOffset f) \
			visGrpRangeOff: (ReadOffset f) \
			visGrpIndxOff: (ReadOffset f) \
			u5: (ReadBElong f) \
		)
		
		fseek f Rtn#seek_set
		if PrintDebug == true do(print (fshp_info))
		append FSHPArr(fshp_info)
	)
	-- F_Model Inverse Matrices --
	--fseek f fskl_info.invMatrArrOff#seek_set
	NxtMatr = 0
	fseek f fskl_info.invIndxArrOff#seek_set
	For indx = 1 to fskl_info.invIndxArrCount Do (
		invIndx = ReadBEShort f +1
		NxtIndx = ftell f
		
		fseek f (fskl_info.invMatrArrOff + NxtMatr)#seek_set
		m11 = ReadBEFloat f; m21 = ReadBEFloat f; m31 = ReadBEFloat f; m41 = ReadBEFloat f
		m12 = ReadBEFloat f; m22 = ReadBEFloat f; m32 = ReadBEFloat f; m42 = ReadBEFloat f
		m13 = ReadBEFloat f; m23 = ReadBEFloat f; m33 = ReadBEFloat f; m43 = ReadBEFloat f
		invtfm = matrix3 [m11,m21,m31,m41] [m12,m22,m32,m42] [m13,m23,m33,m43] [0,0,0,1]
		
		NxtMatr += 48
		append invMatrArr (invMatrix invIndx:invIndx invMatr:invtfm)
		fseek f NxtIndx#seek_set
	)

	-- F_Model Bone Creation --
	fseek f fskl_info.boneArrOff#seek_set
	case of (
		(verNum == 0):(
			For bn = 1 to fskl_info.boneArrCount Do (
				boneData=( bdata \
					bNameOff: (ReadOffset f) \
					bIndx: (ReadBEShort f +1) \
					parIndx1: (ReadBEShort f +1) \
					parIndx2: (ReadBEShort f +1) \
					parIndx3: (ReadBEShort f +1) \
					parIndx4: (ReadBEShort f +1) \
					u1: (ReadBEShort f) \
					bFlags: (ReadBEShort f) \
					u2: (ReadBEShort f) \
					scaleX: (ReadBEFloat f) \
					scaleY: (ReadBEFloat f) \
					scaleZ: (ReadBEFloat f) \
					rotX: (ReadBEFloat f) \
					rotY: (ReadBEFloat f) \
					rotZ: (ReadBEFloat f) \
					rotW: (ReadBEFloat f) \
					posX: (ReadBEFloat f) \
					posY: (ReadBEFloat f) \
					posZ: (ReadBEFloat f) \
					padding: (ReadLong f) \
				)
				
				if boneData.u2 == 0x0001 then(
				tfm = (quat boneData.rotX boneData.rotY boneData.rotZ boneData.rotW) as matrix3
				tfm = inverse(tfm)
				) else (tfm = (eulerangles (boneData.rotX*(180/pi)) (boneData.rotY*(180/pi)) (boneData.rotZ*(180/pi))) as matrix3)
				tfm.row4 = [boneData.posX,boneData.posY,boneData.posZ]
				
				newBone = bonesys.createbone \
					tfm.row4	\
					(tfm.row4 + 0.01 * (normalize tfm.row1)) \
					(normalize tfm.row3)
					
				Rtn = ftell f
				fseek f boneData.bNameOff#seek_set
				boneName = ReadString f
				fseek f Rtn#seek_set
				newBone.name = boneName
				newBone.width  = 0.03
				newBone.height = 0.03
				newBone.transform = tfm
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()
				
				If boneData.parIndx1 != 65536 Do (
					newBone.parent = BoneArray[boneData.parIndx1]
					newBone.transform *= BoneArray[boneData.parIndx1].transform
				)
				
				append BoneArray newBone
				append BoneFixArray newBone.transform
				fseek f 48#seek_cur
			)
		)
		(verNum == 2):(
			For bn = 1 to fskl_info.boneArrCount Do (
				boneData=( bdata \
					bNameOff: (ReadOffset f) \
					bIndx: (ReadBEShort f +1) \
					parIndx1: (ReadBEShort f +1) \
					parIndx2: (ReadBEShort f +1) \
					parIndx3: (ReadBEShort f +1) \
					parIndx4: (ReadBEShort f +1) \
					u1: (ReadBEShort f) \
					bFlags: (ReadBEShort f) \
					u2: (ReadBEShort f) \
					scaleX: (ReadBEFloat f) \
					scaleY: (ReadBEFloat f) \
					scaleZ: (ReadBEFloat f) \
					rotX: (ReadBEFloat f) \
					rotY: (ReadBEFloat f) \
					rotZ: (ReadBEFloat f) \
					rotW: (ReadBEFloat f) \
					posX: (ReadBEFloat f) \
					posY: (ReadBEFloat f) \
					posZ: (ReadBEFloat f) \
					padding: (ReadLong f) \
				)
				
				if boneData.u2 == 0x0001 then(
				tfm = (quat boneData.rotX boneData.rotY boneData.rotZ boneData.rotW) as matrix3
				tfm = inverse(tfm)
				) else (tfm = (eulerangles (boneData.rotX*(180/pi)) (boneData.rotY*(180/pi)) (boneData.rotZ*(180/pi))) as matrix3)
				tfm.row4 = [boneData.posX,boneData.posY,boneData.posZ]
				
				newBone = bonesys.createbone \
					tfm.row4	\
					(tfm.row4 + 0.01 * (normalize tfm.row1)) \
					(normalize tfm.row3)
					
				Rtn = ftell f
				fseek f boneData.bNameOff#seek_set
				boneName = ReadString f
				fseek f Rtn#seek_set
				newBone.name = boneName
				newBone.width  = 0.03
				newBone.height = 0.03
				newBone.transform = tfm
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()
				
				If boneData.parIndx1 != 65536 Do (
					newBone.parent = BoneArray[boneData.parIndx1]
					newBone.transform *= BoneArray[boneData.parIndx1].transform
				)
				
				append BoneArray newBone
				append BoneFixArray newBone.transform
				fseek f 48#seek_cur
			)
		)
		(verNum == 3):(
			For bn = 1 to fskl_info.boneArrCount Do (
				boneData=( bdata \
					bNameOff: (ReadOffset f) \
					bIndx: (ReadBEShort f +1) \
					parIndx1: (ReadBEShort f +1) \
					parIndx2: (ReadBEShort f +1) \
					parIndx3: (ReadBEShort f +1) \
					parIndx4: (ReadBEShort f +1) \
					u1: (ReadBEShort f) \
					bFlags: (ReadBEShort f) \
					u2: (ReadBEShort f) \
					scaleX: (ReadBEFloat f) \
					scaleY: (ReadBEFloat f) \
					scaleZ: (ReadBEFloat f) \
					rotX: (ReadBEFloat f) \
					rotY: (ReadBEFloat f) \
					rotZ: (ReadBEFloat f) \
					rotW: (ReadBEFloat f) \
					posX: (ReadBEFloat f) \
					posY: (ReadBEFloat f) \
					posZ: (ReadBEFloat f) \
					padding: (ReadLong f) \
				)
				
				if boneData.u2 == 0x0001 then(
				tfm = (quat boneData.rotX boneData.rotY boneData.rotZ boneData.rotW) as matrix3
				tfm = inverse(tfm)
				) else (tfm = (eulerangles (boneData.rotX*(180/pi)) (boneData.rotY*(180/pi)) (boneData.rotZ*(180/pi))) as matrix3)
				tfm.row4 = [boneData.posX,boneData.posY,boneData.posZ]
				
				newBone = bonesys.createbone \
					tfm.row4	\
					(tfm.row4 + 0.01 * (normalize tfm.row1)) \
					(normalize tfm.row3)
					
				Rtn = ftell f
				fseek f boneData.bNameOff#seek_set
				boneName = ReadString f
				fseek f Rtn#seek_set
				newBone.name = boneName
				newBone.width  = 0.03
				newBone.height = 0.03
				newBone.transform = tfm
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()
				
				If boneData.parIndx1 != 65536 Do (
					newBone.parent = BoneArray[boneData.parIndx1]
					newBone.transform *= BoneArray[boneData.parIndx1].transform
				)
				
				append BoneArray newBone
				append BoneFixArray newBone.transform
				fseek f 48#seek_cur
			)
		)
		(verNum == 4):(
			For bn = 1 to fskl_info.boneArrCount Do (
				boneData=( bdata \
					bNameOff: (ReadOffset f) \
					bIndx: (ReadBEShort f +1) \
					parIndx1: (ReadBEShort f +1) \
					parIndx2: (ReadBEShort f +1) \
					parIndx3: (ReadBEShort f +1) \
					parIndx4: (ReadBEShort f +1) \
					u1: (ReadBEShort f) \
					bFlags: (ReadBEShort f) \
					u2: (ReadBEShort f) \
					scaleX: (ReadBEFloat f) \
					scaleY: (ReadBEFloat f) \
					scaleZ: (ReadBEFloat f) \
					rotX: (ReadBEFloat f) \
					rotY: (ReadBEFloat f) \
					rotZ: (ReadBEFloat f) \
					rotW: (ReadBEFloat f) \
					posX: (ReadBEFloat f) \
					posY: (ReadBEFloat f) \
					posZ: (ReadBEFloat f) \
					padding: (ReadLong f) \
				)
				
				if boneData.u2 == 0x0001 then(
				tfm = (quat boneData.rotX boneData.rotY boneData.rotZ boneData.rotW) as matrix3
				tfm = inverse(tfm)
				) else (tfm = (eulerangles (boneData.rotX*(180/pi)) (boneData.rotY*(180/pi)) (boneData.rotZ*(180/pi))) as matrix3)
				tfm.row4 = [boneData.posX,boneData.posY,boneData.posZ]
				
				newBone = bonesys.createbone \
					tfm.row4	\
					(tfm.row4 + 0.01 * (normalize tfm.row1)) \
					(normalize tfm.row3)
					
				Rtn = ftell f
				fseek f boneData.bNameOff#seek_set
				boneName = ReadString f
				fseek f Rtn#seek_set
				newBone.name = boneName
				newBone.width  = 0.03
				newBone.height = 0.03
				newBone.transform = tfm
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()
				
				If boneData.parIndx1 != 65536 Do (
					newBone.parent = BoneArray[boneData.parIndx1]
					newBone.transform *= BoneArray[boneData.parIndx1].transform
				)
				
				append BoneArray newBone
				append BoneFixArray newBone.transform
			)
		)
		(verNum == 5):(
			For bn = 1 to fskl_info.boneArrCount Do (
				boneData=( bdata \
					bNameOff: (ReadOffset f) \
					bIndx: (ReadBEShort f +1) \
					parIndx1: (ReadBEShort f +1) \
					parIndx2: (ReadBEShort f +1) \
					parIndx3: (ReadBEShort f +1) \
					parIndx4: (ReadBEShort f +1) \
					u1: (ReadBEShort f) \
					bFlags: (ReadBEShort f) \
					u2: (ReadBEShort f) \
					scaleX: (ReadBEFloat f) \
					scaleY: (ReadBEFloat f) \
					scaleZ: (ReadBEFloat f) \
					rotX: (ReadBEFloat f) \
					rotY: (ReadBEFloat f) \
					rotZ: (ReadBEFloat f) \
					rotW: (ReadBEFloat f) \
					posX: (ReadBEFloat f) \
					posY: (ReadBEFloat f) \
					posZ: (ReadBEFloat f) \
					padding: (ReadLong f) \
				)
			
				if boneData.u2 == 0x0001 then(
				tfm = (quat boneData.rotX boneData.rotY boneData.rotZ boneData.rotW) as matrix3
				tfm = inverse(tfm)
				) else (tfm = (eulerangles (boneData.rotX*(180/pi)) (boneData.rotY*(180/pi)) (boneData.rotZ*(180/pi))) as matrix3)
				tfm.row4 = [boneData.posX,boneData.posY,boneData.posZ]
				
				newBone = bonesys.createbone \
					tfm.row4	\
					(tfm.row4 + 0.01 * (normalize tfm.row1)) \
					(normalize tfm.row3)
					
				Rtn = ftell f
				fseek f boneData.bNameOff#seek_set
				boneName = ReadString f
				fseek f Rtn#seek_set
				newBone.name = boneName
				newBone.width  = 0.03
				newBone.height = 0.03
				newBone.transform = tfm
				newBone.setBoneEnable false 0
				newBone.wirecolor = yellow
				newbone.showlinks = true
				newBone.pos.controller      = TCB_position ()
				newBone.rotation.controller = TCB_rotation ()
				
				If boneData.parIndx1 != 65536 Do (
					newBone.parent = BoneArray[boneData.parIndx1]
					newBone.transform *= BoneArray[boneData.parIndx1].transform
				)
				
				append BoneArray newBone
				append BoneFixArray newBone.transform
			)
		)
	)

	-- Mesh Building --
	For m = 1 to FSHPArr.count Do (
		Vert_Array = #()
		Normal_Array = #()
		UV_Array = #()
		UV2_Array = #()
		UV3_Array = #()
		UV4_Array = #()
		UV5_Array = #()
		Color_Array = #()
		AttrArr = #()
		BuffArr = #()
		B1_array = #()
		B2_array = #()
		W1_array = #()
		W2_array = #()
		Weight_array = #()
		
		fseek f FSHPArr[m].polyNameOff#seek_set
		MeshName = ReadString f

		fseek f FVTXArr[FSHPArr[m].fvtxIndx].attArrOff#seek_set
		if PrintDebug == true do(print ("Building polygon " + MeshName as string + "..."))
		For att = 1 to FVTXArr[FSHPArr[m].fvtxIndx].attCount Do (
			AttTypeOff = ReadOffset f
			Rtn1 = ftell f
			fseek f AttTypeOff#seek_set
			AttType = ReadString f
			fseek f Rtn1#seek_set
			buffIndx = ReadByte f+1
			skip = ReadByte f
			buffOff = ReadBEShort f
			vertType = ReadBElong f
			append AttrArr (attdata attName: AttType buffIndx: buffIndx buffOff: buffOff vertType: vertType)
		)
		if PrintDebug == true do(print AttrArr)
		
		fseek f FVTXArr[FSHPArr[m].fvtxIndx].buffArrOff#seek_set
		For buf = 1 to FVTXArr[FSHPArr[m].fvtxIndx].buffCount Do (
			unk1 = ReadLong f
			BufferSize = ReadBELong f
			unk2 = ReadLong f
			StrideSize = ReadBEShort f
			unk3 = ReadShort f
			unk4 = ReadLong f
			DataOffset = ReadOffset f
			append BuffArr (buffData buffSize: BufferSize strideSize: StrideSize dataOffset: DataOffset)
		)

		fseek f BuffArr[1].dataOffset #seek_set
		For attr = 1 to AttrArr.count Do (
			locPoint = ((BuffArr[AttrArr[attr].buffIndx].dataOffset) + (AttrArr[attr].buffOff))
			fseek f locPoint #seek_set
			For v = 1 to FVTXArr[FSHPArr[m].fvtxIndx].vertCount Do (
				VertStart = (ftell f + (BuffArr[AttrArr[attr].buffIndx].strideSize))
				case of (
					-- 0x00 - Unsigned ints
					default: (throw("Unknown type!"))
					(AttrArr[attr].vertType == 0x0004):(
						-- 2 byte integers, unsigned
						var1 = ReadByte f #unsigned
						var2 = ReadByte f #unsigned
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x0007):(
						-- 2 short integers, unsigned
						var1 = ReadBEShort f
						var2 = ReadBEShort f
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x000A):(
					-- 4 byte integers, unsigned
						var1 = ReadByte f #unsigned
						var2 = ReadByte f #unsigned
						var3 = ReadByte f #unsigned
						var4 = ReadByte f #unsigned
					)

					-- 0x01 - Unsigned ints (bone IDs only?)
					(AttrArr[attr].vertType == 0x0100):(
						-- 1 byte integer, unsigned, bone IDs
						var1 = ReadByte f #unsigned + 1
						var2 = 1
						var3 = 1
						var4 = 1
					)
					(AttrArr[attr].vertType == 0x0104):(
						-- 2 byte integers, unsigned, bone IDs
						var1 = ReadByte f #unsigned + 1
						var2 = ReadByte f #unsigned + 1
						var3 = 1
						var4 = 1
					)
					(AttrArr[attr].vertType == 0x010A):(
						-- 4 byte integers, unsigned, bone IDs
						var1 = ReadByte f #unsigned + 1
						var2 = ReadByte f #unsigned + 1
						var3 = ReadByte f #unsigned + 1
						var4 = ReadByte f #unsigned + 1
					)
					(AttrArr[attr].vertType == 0x010C):(
						-- 2 long integers, unsigned, bone IDs
						var1 = ReadLong f #unsigned + 1
						var2 = ReadLong f #unsigned + 1
						var3 = 1
						var4 = 1
					)
					(AttrArr[attr].vertType == 0x0110):(
						-- 4 long integers, unsigned, bone IDs
						var1 = ReadLong f #unsigned + 1
						var2 = ReadLong f #unsigned + 1
						var3 = ReadLong f #unsigned + 1
						var4 = ReadLong f #unsigned + 1
					)
					(AttrArr[attr].vertType == 0x0112):(
						-- 4 long integers, unsigned, bone IDs, same?
						var1 = ReadLong f #unsigned + 1
						var2 = ReadLong f #unsigned + 1
						var3 = ReadLong f #unsigned + 1
						var4 = ReadLong f #unsigned + 1
					)

					-- 0x02 - Signed ints
					(AttrArr[attr].vertType == 0x0204):(
						-- 2 byte integers, signed
						var1 = ReadByte f
						var2 = ReadByte f
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x0207):(
						-- 2 short integers, signed
						var1 = ReadBEShort f
						var2 = ReadBEShort f
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x020A):(
						-- 4 byte integers, signed
						var1 = ReadByte f
						var2 = ReadByte f
						var3 = ReadByte f
						var4 = ReadByte f
					)
					(AttrArr[attr].vertType == 0x020B):(
						-- 3 10-bit values, signed
						vars = ReadBElong f
						var1 = sign10(bit.and (vars) 0x3FF) as float / 511
						var2 = sign10(bit.and (bit.shift vars -10) 0x3FF) as float / 511
						var3 = sign10(bit.and (bit.shift vars -20) 0x3FF) as float / 511
						var4 = 0
					)

					-- 0x08 - Floats
					(AttrArr[attr].vertType == 0x0808):(
						-- 2 half-floats
						var1 = readBEHalfFloat f
						var2 = readBEHalfFloat f
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x080D):(
						-- 2 floats
						var1 = ReadBEFloat f
						var2 = ReadBEFloat f
						var3 = 0
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x080F):(
						-- 4 half-floats
						var1 = ReadBEHalfFloat f
						var2 = ReadBEHalfFloat f
						var3 = ReadBEHalfFloat f
						var4 = ReadBEHalfFloat f
					)
					(AttrArr[attr].vertType == 0x0811):(
						-- 3 floats
						var1 = ReadBEFloat f
						var2 = ReadBEFloat f
						var3 = ReadBEFloat f
						var4 = 0
					)
					(AttrArr[attr].vertType == 0x0813):(
						-- 4 floats
						var1 = ReadBEFloat f
						var2 = ReadBEFloat f
						var3 = ReadBEFloat f
						var4 = ReadBEFloat f
					)
				)

				case of (
					--Vertex Info--
					(AttrArr[attr].attName == "_p0"):(
						append vert_Array[var1,var2,var3]
					)
                    --Color Info--
                    (AttrArr[attr].attName == "_c0"):(
                        case of (
                            (AttrArr[attr].vertType == 0x020A):(
                                var1 = vRange 0 254 (var1 * 2)
                                var2 = vRange 0 254 (var2 * 2)
                                var3 = vRange 0 254 (var3 * 2)
                                var4 = vRange 0 1 (var4 as float / 127)
                            )
                            (AttrArr[attr].vertType == 0x080F):(
                                var1 = vRange 0 255 (var1 * 255)
                                var2 = vRange 0 255 (var2 * 255)
                                var3 = vRange 0 255 (var3 * 255)
                            )
                            (AttrArr[attr].vertType == 0x0813):(
                                var1 = vRange 0 255 (var1 * 255)
                                var2 = vRange 0 255 (var2 * 255)
                                var3 = vRange 0 255 (var3 * 255)
                            )
                        )
                        append Color_Array[var1,var2,var3]
                    )
					--Normals Info--
					(AttrArr[attr].attName == "_n0"):(
						case of (
							(AttrArr[attr].vertType == 0x020A):(
								fseek f (AttrArr[attr].buffOff)#seek_cur
								var1 = var1 as float / 127
								var2 = var2 as float / 127
								var3 = var3 as float / 127
								var4 = var4 as float / 127
							)
						)
						append Normal_Array[var1,var2,var3]
					)
					--Tangent Info--
					--Binormals Info--
					--UV Info--
					(AttrArr[attr].attName == "detail"):(
						var1 *= 2
						var2 *= 2
						append UV3_array [var1,var2,0]
					)
					(AttrArr[attr].attName == "_u0" or AttrArr[attr].attName == "color" or AttrArr[attr].attName == "_u1" or AttrArr[attr].attName == "_u2" or AttrArr[attr].attName == "_u3" or AttrArr[attr].attName == "_u4" or AttrArr[attr].attName == "All"):(
						case of (
							(AttrArr[attr].vertType == 0x0004):(
								var1 = (var1 as float / 255)
								var2 = ((var2 as float / 255) * -1) + 1
							)
							(AttrArr[attr].vertType == 0x0007):(
								var1 = (var1 as float / 65535)
								var2 = ((var2 as float / 65535) * -1) + 1
							)
							(AttrArr[attr].vertType == 0x0204):(
								var1 = (var1 as float / 127)
								var2 = ((var2 as float / 127) * -1) + 1
							)
							(AttrArr[attr].vertType == 0x0207):(
								var1 = (var1 as float / 32767)
								var2 = ((var2 as float / 32767) * -1) + 1
								case of (
									(var1 < -1): var1 += 2
									(var1 > 1): var1 -= 2
								)
								case of (
									(var2 < -1): var2 += 2
									(var2 > 1): var2 -= 2
								)
							)
							(AttrArr[attr].vertType == 0x0808):(
								var2 = (var2 * -1) + 1
							)
							(AttrArr[attr].vertType == 0x080D):(
								var2 = (var2 * -1) + 1
							)
						)
						case of(
							(AttrArr[attr].attName == "_u0" or AttrArr[attr].attName == "color"): append UV_Array[var1,var2,0]
							(AttrArr[attr].attName == "_u1"): append UV2_Array[var1,var2,0]
							(AttrArr[attr].attName == "_u2"): append UV3_Array[var1,var2,0]
							(AttrArr[attr].attName == "_u3"): append UV4_Array[var1,var2,0]
							(AttrArr[attr].attName == "_u4" or AttrArr[attr].attName == "All"): append UV5_Array[var1,var2,0]
						)
					)
						
-- 					(AttrArr[attr].attName == "_u0" or AttrArr[attr].attName == "color"):(
-- 						case of (
-- 							(AttrArr[attr].vertType == 0x0004):(
-- 								var1 = (var1 as float / 255)
-- 								var2 = ((var2 as float / 255) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0007):(
-- 								var1 = (var1 as float / 65535)
-- 								var2 = ((var2 as float / 65535) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0204):(
-- 								var1 = (var1 as float / 127)
-- 								var2 = ((var2 as float / 127) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0207):(
-- 								var1 = (var1 as float / 32767)
-- 								var2 = ((var2 as float / 32767) * -1) + 1
-- 								case of (
-- 									(var1 < -1):(var1 = var1 + 2)
-- 									(var1 > 1):(var1 = var1 - 2)
-- 								)
-- 								case of (
-- 									(var2 < -1):(var2 = var2 + 2)
-- 									(var2 > 1):(var2 = var2 - 2)
-- 								)
-- 							)
-- 							(AttrArr[attr].vertType == 0x0808):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x080D):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 						)
-- 						append UV_Array[var1,var2,0]
-- 					)
-- 					(AttrArr[attr].attName == "_u1"):(
-- 						case of (
-- 							(AttrArr[attr].vertType == 0x0004):(
-- 							var1 = (var1 as float / 255)
-- 							var2 = ((var2 as float / 255) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0007):(
-- 							var1 = (var1 as float / 65535)
-- 							var2 = ((var2 as float / 65535) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0204):(
-- 							var1 = (var1 as float / 127)
-- 							var2 = ((var2 as float / 127) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0207):(
-- 							var1 = (var1 as float / 32767)
-- 							var2 = ((var2 as float / 32767) * -1) + 1
-- 							case of (
-- 								(var1 < -1):(var1 = var1 + 2)
-- 								(var1 > 1):(var1 = var1 - 2)
-- 							)
-- 							case of (
-- 								(var2 < -1):(var2 = var2 + 2)
-- 								(var2 > 1):(var2 = var2 - 2)
-- 							)
-- 							)
-- 							(AttrArr[attr].vertType == 0x0808):(
-- 							var2 = (var2 * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x080D):(
-- 							var2 = (var2 * -1) + 1
-- 							)
-- 						)
-- 						append UV2_Array[var1,var2,0]
-- 					)
-- 					(AttrArr[attr].attName == "_u2"):(
-- 						case of (
-- 							(AttrArr[attr].vertType == 0x0004):(
-- 								var1 = (var1 as float / 255)
-- 								var2 = ((var2 as float / 255) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0007):(
-- 								var1 = (var1 as float / 65535)
-- 								var2 = ((var2 as float / 65535) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0204):(
-- 								var1 = (var1 as float / 127)
-- 								var2 = ((var2 as float / 127) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0207):(
-- 								var1 = (var1 as float / 32767)
-- 								var2 = ((var2 as float / 32767) * -1) + 1
-- 								case of (
-- 									(var1 < -1):(var1 = var1 + 2)
-- 									(var1 > 1):(var1 = var1 - 2)
-- 								)
-- 								case of (
-- 									(var2 < -1):(var2 = var2 + 2)
-- 									(var2 > 1):(var2 = var2 - 2)
-- 								)
-- 							)
-- 							(AttrArr[attr].vertType == 0x0808):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x080D):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 						)
-- 						append UV3_Array[var1,var2,0]
-- 					)
-- 					(AttrArr[attr].attName == "_u3"):(
-- 						case of (
-- 							(AttrArr[attr].vertType == 0x0004):(
-- 								var1 = (var1 as float / 255)
-- 								var2 = ((var2 as float / 255) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0007):(
-- 								var1 = (var1 as float / 65535)
-- 								var2 = ((var2 as float / 65535) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0204):(
-- 								var1 = (var1 as float / 127)
-- 								var2 = ((var2 as float / 127) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0207):(
-- 								var1 = (var1 as float / 32767)
-- 								var2 = ((var2 as float / 32767) * -1) + 1
-- 								case of (
-- 									(var1 < -1):(var1 = var1 + 2)
-- 									(var1 > 1):(var1 = var1 - 2)
-- 								)
-- 								case of (
-- 									(var2 < -1):(var2 = var2 + 2)
-- 									(var2 > 1):(var2 = var2 - 2)
-- 								)
-- 							)
-- 							(AttrArr[attr].vertType == 0x0808):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x080D):(
-- 								var2 = (var2 * -1) + 1
-- 							)
-- 						)
-- 						append UV4_Array[var1,var2,0]
-- 					)
-- 					(AttrArr[attr].attName == "_u4" or AttrArr[attr].attName == "All"):(
-- 						case of (
-- 							(AttrArr[attr].vertType == 0x0004):(
-- 							var1 = (var1 as float / 255)
-- 							var2 = ((var2 as float / 255) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0007):(
-- 							var1 = (var1 as float / 65535)
-- 							var2 = ((var2 as float / 65535) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0204):(
-- 							var1 = (var1 as float / 127)
-- 							var2 = ((var2 as float / 127) * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x0207):(
-- 							var1 = (var1 as float / 32767)
-- 							var2 = ((var2 as float / 32767) * -1) + 1
-- 							case of (
-- 								(var1 < -1):(var1 = var1 + 2)
-- 								(var1 > 1):(var1 = var1 - 2)
-- 							)
-- 							case of (
-- 								(var2 < -1):(var2 = var2 + 2)
-- 								(var2 > 1):(var2 = var2 - 2)
-- 							)
-- 							)
-- 							(AttrArr[attr].vertType == 0x0808):(
-- 							var2 = (var2 * -1) + 1
-- 							)
-- 							(AttrArr[attr].vertType == 0x080D):(
-- 							var2 = (var2 * -1) + 1
-- 							)
-- 						)
-- 						append UV5_Array[var1,var2,0]
-- 					)
-- 					(AttrArr[attr].attName == "detail"):(
-- 						var1 = var1 * 2
-- 						var2 = var2 * 2
-- 						append UV3_array [var1,var2,0]
-- 					)

					--BlendIndex Info--
					(AttrArr[attr].attName == "_i0"):(
						bone1 = Node_array[var1]
						bone2 = Node_array[var2]
						bone3 = Node_array[var3]
						bone4 = Node_array[var4]
						append B1_Array (Bone_Info_Struct bone1:bone1 bone2:bone2 bone3:bone3 bone4:bone4)
						if AttrArr[attr].vertType == 256 do(append W1_Array (Weight_Info_Struct weight1:1 weight2:0 weight3:0 weight4:0))
					)
					(AttrArr[attr].attName == "_i1"):(
						bone5 = Node_array[var1]
						bone6 = Node_array[var2]
						bone7 = Node_array[var3]
						bone8 = Node_array[var4]
						append B2_Array (Bone_Info_Struct bone1:bone5 bone2:bone6 bone3:bone7 bone4:bone8)
					)
					-- BlendWeight Info --
					(AttrArr[attr].attName == "_w0"):(
						case of (
							(AttrArr[attr].vertType == 0x0004):(
							weight1 = var1 as float / 255
							weight2 = var2 as float / 255
							weight3 = 0
							weight4 = 0
							)
							(AttrArr[attr].vertType == 0x000A):(
							weight1 = var1 as float / 255
							weight2 = var2 as float / 255
							weight3 = var3 as float / 255
							weight4 = var4 as float / 255
							if weight4 >= 1 do(weight4 = 0)
							)
						)
						append W1_Array (Weight_Info_Struct weight1:weight1 weight2:weight2 weight3:weight3 weight4:weight4)
					)
					(AttrArr[attr].attName == "_w1"):( 
						case of (
							(AttrArr[attr].vertType == 0x0004):(
							weight5 = var1 as float / 255
							weight6 = var2 as float / 255
							weight7 = 0
							weight8 = 0
							)
							(AttrArr[attr].vertType == 0x000A):(
							weight5 = var1 as float / 255
							weight6 = var2 as float / 255
							weight7 = var3 as float / 255
							weight8 = var4 as float / 255
							if weight8 >= 1 do(weight8 = 0)
							)
						)
						append W2_Array (Weight_Info_Struct weight1:weight5 weight2:weight6 weight3:weight7 weight4:weight8)
					)
				)

				fseek f VertStart #seek_set
			) -- Close Vert Read Loop
		) -- Close Attributes Loop

		If Color_Array.count == 0 Do (
			For ca = 1 to Vert_array.count Do (
				append Color_Array[255,255,255]
			)
		)

		If B2_Array.count == 0 Do (
			For v = 1 to B1_array.count Do (
				append B2_Array (Bone_Info_Struct bone1:1 bone2:1 bone3:1 bone4:1)
				append W2_Array (Weight_Info_Struct weight1:0 weight2:0 weight3:0 weight4:0)
			)
		)

		for b = 1 to W1_Array.count Do (
			w = (weight_data boneids:#() weights:#())
			maxweight = 0
			
			if(W1_array[b].Weight1 != 0) then
				maxweight = maxweight + W1_array[b].Weight1
			if(W1_array[b].Weight2 != 0) then
				maxweight = maxweight + W1_array[b].Weight2
			if(W1_array[b].Weight3 != 0) then
				maxweight = maxweight + W1_array[b].Weight3
			if(W1_array[b].Weight4 != 0) then
				maxweight = maxweight + W1_array[b].Weight4
			if(W2_array[b].Weight1 != 0) then
				maxweight = maxweight + W2_array[b].Weight1
			if(W2_array[b].Weight2 != 0) then
				maxweight = maxweight + W2_array[b].Weight2
			if(W2_array[b].Weight3 != 0) then
				maxweight = maxweight + W2_array[b].Weight3
			if(W2_array[b].Weight4 != 0) then
				maxweight = maxweight + W2_array[b].Weight4
				
			if(maxweight != 0) then
			(
				if(W1_array[b].Weight1 != 0) then
				(
					w1 = W1_array[b].Weight1 as float
					append w.boneids (B1_array[b].Bone1)
					append w.weights (w1)
				)
				if(W1_array[b].Weight2 != 0) then
				(
					w2 = W1_array[b].Weight2 as float
					append w.boneids (B1_array[b].Bone2)
					append w.weights (w2)
				)
				if(W1_array[b].Weight3 != 0) then
				(
					w3 = W1_array[b].Weight3 as float
					append w.boneids (B1_array[b].Bone3)
					append w.weights (w3)
				)
				if(W1_array[b].Weight4 != 0) then
				(
					w4 = W1_array[b].Weight4 as float
					append w.boneids (B1_array[b].Bone4)
					append w.weights (w4)
				)
				if(W2_array[b].Weight1 != 0) then
				(
					w5 = W2_array[b].Weight1 as float
					append w.boneids (B2_array[b].Bone1)
					append w.weights (w5)
				)
				if(W2_array[b].Weight2 != 0) then
				(
					w6 = W2_array[b].Weight2 as float
					append w.boneids (B2_array[b].Bone2)
					append w.weights (w6)
				)
				if(W2_array[b].Weight3 != 0) then
				(
					w7 = W2_array[b].Weight3 as float
					append w.boneids (B2_array[b].Bone3)
					append w.weights (w7)
				)
				if(W2_array[b].Weight4 != 0) then
				(
					w8 = W2_array[b].Weight4 as float
					append w.boneids (B2_array[b].Bone4)
					append w.weights (w8)
				)
			)
			append Weight_array w
		)

		fseek f FSHPArr[m].lodMdlOff#seek_set
		if LODs == 1 do(FSHPArr[m].lodMdlCount =  1) 
		for l = 1 to FSHPArr[m].lodMdlCount do(
			Face_Array = #()
			lodmdl_info=(lodmdl \
				u1: (ReadBELong f) \
				faceType: (ReadBELong f) \
				dCount: (ReadBELong f) \
				visGrpCount: (ReadBEShort f) \
				u3: (ReadBEShort f) \
				visGrpOff: (ReadOffset f) \
				indxBuffOff: (ReadOffset f) \
				elmSkip: (ReadBELong f) \
			)
			LODRet = (ftell f)

			fseek f lodmdl_info.indxBuffOff#seek_set
			unk1 = ReadBELong f
			FaceCount = (ReadBElong f)
			fseek f 12#seek_cur
			FaceBuffer = ReadOffset f
			
			fseek f FaceBuffer#seek_set
			PolyStart = lodmdl_info.elmSkip
			case of (
				(lodmdl_info.faceType == 4):(
					FaceCount = (FaceCount/6)
					For face = 1 to FaceCount Do (
					fa = ((ReadBEShort f) + 1 + PolyStart)
					fb = ((ReadBEShort f) + 1 + PolyStart)
					fc = ((ReadBEShort f) + 1 + PolyStart)
					append Face_Array[fa,fb,fc]
					)
				)
				(lodmdl_info.faceType == 9):(
					FaceCount = (FaceCount/12)
					For face = 1 to FaceCount Do (
						fa = ((ReadBELong f) + 1 + PolyStart)
						fb = ((ReadBELong f) + 1 + PolyStart)
						fc = ((ReadBELong f) + 1 + PolyStart)
						append Face_Array[fa,fb,fc]
					)
				)
			)

			BoneCount = BoneArray.count
			local PolyMat = standardMaterial()
			PolyMat.name = (FMATNameArr[FSHPArr[m].fmatIndx] as string)
			PolyMat.showinviewport = true
			PolyMat.twosided = false
			fseek f (FMATArr[FSHPArr[m].fmatIndx].texSelOff)#seek_set
			TextureNameOffset = readoffset f
			fseek f TextureNameOffset#seek_set
			TextureName = readstring f
			if TexFormat == 2 then (tm = Bitmaptexture filename:(p + TextureName as string + ".png")) else (tm = Bitmaptexture filename:(p + TextureName as string + ".dds"))
			tm.alphasource = 0
			PolyMat.diffuseMap = tm
			PolyMat.opacityMap = tm
			PolyMat.opacityMap.monoOutput = 1

			msh = mesh vertices:Vert_array faces:Face_array
			msh.numTVerts = Vert_Array.count
			setNumCPVVerts msh msh.numTVerts
			setCVertMode msh true
			setShadeCVerts msh true
			defaultVCFaces msh
			buildTVFaces msh
			for j = 1 to Color_Array.count do setvertcolor msh j Color_array[j]

			If Vert_Array.count == UV_Array.count Do (
				for j = 1 to Vert_array.count do setTVert msh j UV_array[j]
				for j = 1 to Face_array.count do setTVFace msh j Face_array[j]
			)
			for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
				
			if FSHPArr[m].lodMdlCount == 1 then(msh.name = (MeshName)) else (msh.name = (MeshName + " (LOD: " + l as string + ")"))
			msh.material = PolyMat
			
			max modify mode
			select msh

			addmodifier msh (Edit_Normals ()) ui:off
			msh.Edit_Normals.MakeExplicit selection:#{1..Normal_array.count}
			EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
			EN_setNormal = msh.Edit_Normals.SetNormal
			normID = #{}
			for v = 1 to Normal_array.count do
			(
				free normID
				EN_convertVS #{v} &normID
				for id in normID do EN_setNormal id Normal_array[v]
			)

			if BoneCount > 0 do(
					if FSHPArr[m].matrFlag == 0 then(
						$.transform = BoneFixArray[FSHPArr[m].fsklIndx]
						if RigModel != 2 do(
						skinMod = skin ()
						addModifier msh skinMod
						maxbone = getnodebyname BoneArray[FSHPArr[m].fsklIndx].name
						skinOps.addBone skinMod maxbone 0
						)
					) else ( if RigModel != 2 do(
					skinMod = skin ()
					boneIDMap = #()
					addModifier msh skinMod
					msh.Skin.weightAllVertices = false

					if FSHPArr[m].matrFlag == 1 do(
						for b in BoneArray do b.transform = matrix3 1
					)

					for i = 1 to BoneCount do
					(
						maxbone = getnodebyname BoneArray[i].name
						if i != BoneCount then (
							skinOps.addBone skinMod maxbone 0
						)else(
							skinOps.addBone skinMod maxbone 1
						)
					)
					
					local numSkinBones = skinOps.GetNumberBones skinMod
					for i = 1 to numSkinBones do
					(
						local boneName = skinOps.GetBoneName skinMod i 0
						for j = 1 to BoneCount do
						(
							if boneName == BoneArray[j].Name then
							(
								boneIDMap[j] = i
								j = BoneCount + 1
							)
						)
					) -- Thanks to sunnydavis for the fix!
					
					modPanel.setCurrentObject skinMod

				for i = 1 to Vert_array.count do(
					skinOps.SetVertexWeights skinMod i 1 1
					skinOps.unnormalizeVertex skinMod i true 
					skinOps.SetVertexWeights skinMod i 1 0
				)
					skinOps.RemoveZeroWeights skinMod
				for i = 1 to Vert_array.count do(
					skinOps.unnormalizeVertex skinMod i false
				)

					for i = 1 to Weight_array.count do (
						w = Weight_array[i]
						bi = #() --bone index array
						wv = #() --weight value array
						
						for j = 1 to w.boneids.count do
						(
							boneid = w.boneids[j]
							weight = w.weights[j]
							append bi boneIDMap[boneid]
							append wv weight
						)   
						skinOps.ReplaceVertexWeights skinMod i bi wv
					)

					if FSHPArr[m].matrFlag == 1 do(
						for b = 1 to BoneCount do BoneArray[b].transform = BoneFixArray[b]
					)

				)
				)
			)

		if UVLayers == 1 do(
		if UV2_array.count > 1 do(
			local PolyMat = standardMaterial()
			PolyMat.name = (FMATNameArr[FSHPArr[m].fmatIndx] as string)
			PolyMat.showinviewport = true
			PolyMat.twosided = false
			fseek f (FMATArr[FSHPArr[m].fmatIndx].texSelOff)#seek_set
			TextureNameOffset = readoffset f
			fseek f TextureNameOffset#seek_set
			TextureName = readstring f
			if TexFormat == 2 then (tm = Bitmaptexture filename:(p + TextureName as string + ".png")) else (tm = Bitmaptexture filename:(p + TextureName as string + ".dds"))
			tm.alphasource = 0
			PolyMat.diffuseMap = tm
			PolyMat.opacityMap = tm
			PolyMat.opacityMap.monoOutput = 1
			
			msh = mesh vertices:Vert_array faces:Face_array
			msh.numTVerts = Vert_Array.count
			setNumCPVVerts msh msh.numTVerts
			setCVertMode msh true
			setShadeCVerts msh true
			defaultVCFaces msh
			buildTVFaces msh
			for j = 1 to Color_Array.count do setvertcolor msh j Color_array[j]
			If Vert_Array.count == UV2_Array.count Do (
				for j = 1 to Vert_array.count do setTVert msh j UV2_array[j]
				for j = 1 to Face_array.count do setTVFace msh j Face_array[j]
			)
			for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
				
			if FSHPArr[m].lodMdlCount == 1 then(msh.name = (MeshName) + " Layer 2") else (msh.name = (MeshName + " Layer 2" + " (LOD: " + l as string + ")"))
			msh.material = PolyMat

			select msh
			
			addmodifier msh (Edit_Normals ()) ui:off
			msh.Edit_Normals.MakeExplicit selection:#{1..Normal_array.count}
			EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
			EN_setNormal = msh.Edit_Normals.SetNormal
			normID = #{}
			for v = 1 to Normal_array.count do
			(
				free normID
				EN_convertVS #{v} &normID
				for id in normID do EN_setNormal id Normal_array[v]
			)

			if BoneCount > 0 do(
					if FSHPArr[m].matrFlag == 0 then(
						$.transform = BoneFixArray[FSHPArr[m].fsklIndx]
						if RigModel != 2 do(
						skinMod = skin ()
						addModifier msh skinMod
						maxbone = getnodebyname BoneArray[FSHPArr[m].fsklIndx].name
						skinOps.addBone skinMod maxbone 0
						)
					) else ( if RigModel != 2 do(
					skinMod = skin ()
					boneIDMap = #()
					addModifier msh skinMod
					msh.Skin.weightAllVertices = false

					if FSHPArr[m].matrFlag == 1 do(
						for b in BoneArray do b.transform = matrix3 1
					)

					for i = 1 to BoneCount do
					(
						maxbone = getnodebyname BoneArray[i].name
						if i != BoneCount then (
							skinOps.addBone skinMod maxbone 0
						)else(
							skinOps.addBone skinMod maxbone 1
						)
					)
					
					local numSkinBones = skinOps.GetNumberBones skinMod
					for i = 1 to numSkinBones do
					(
						local boneName = skinOps.GetBoneName skinMod i 0
						for j = 1 to BoneCount do
						(
							if boneName == BoneArray[j].Name then
							(
								boneIDMap[j] = i
								j = BoneCount + 1
							)
						)
					) -- Thanks to sunnydavis for the fix!
					
					modPanel.setCurrentObject skinMod

				for i = 1 to Vert_array.count do(
					skinOps.SetVertexWeights skinMod i 1 1
					skinOps.unnormalizeVertex skinMod i true 
					skinOps.SetVertexWeights skinMod i 1 0
				)
					skinOps.RemoveZeroWeights skinMod
				for i = 1 to Vert_array.count do(
					skinOps.unnormalizeVertex skinMod i false
				)

					for i = 1 to Weight_array.count do (
						w = Weight_array[i]
						bi = #() --bone index array
						wv = #() --weight value array
						
						for j = 1 to w.boneids.count do
						(
							boneid = w.boneids[j]
							weight = w.weights[j]
							append bi boneIDMap[boneid]
							append wv weight
						)   
						skinOps.ReplaceVertexWeights skinMod i bi wv
					)

					if FSHPArr[m].matrFlag == 1 do(
						for b = 1 to BoneCount do BoneArray[b].transform = BoneFixArray[b]
					)

				)
				)
			)

		)

		if UV3_array.count > 1 do(
			local PolyMat = standardMaterial()
			PolyMat.name = (FMATNameArr[FSHPArr[m].fmatIndx] as string)
			PolyMat.showinviewport = true
			PolyMat.twosided = false
			fseek f (FMATArr[FSHPArr[m].fmatIndx].texSelOff)#seek_set
			TextureNameOffset = readoffset f
			fseek f TextureNameOffset#seek_set
			TextureName = readstring f
			if TexFormat == 2 then (tm = Bitmaptexture filename:(p + TextureName as string + ".png")) else (tm = Bitmaptexture filename:(p + TextureName as string + ".dds"))
			tm.alphasource = 0
			PolyMat.diffuseMap = tm
			PolyMat.opacityMap = tm
			PolyMat.opacityMap.monoOutput = 1
			
			msh = mesh vertices:Vert_array faces:Face_array
			msh.numTVerts = Vert_Array.count
			setNumCPVVerts msh msh.numTVerts
			setCVertMode msh true
			setShadeCVerts msh true
			defaultVCFaces msh
			buildTVFaces msh
			for j = 1 to Color_Array.count do setvertcolor msh j Color_array[j]
			If Vert_Array.count == UV3_Array.count Do (
				for j = 1 to Vert_array.count do setTVert msh j UV3_array[j]
				for j = 1 to Face_array.count do setTVFace msh j Face_array[j]
			)
			for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
				
			if FSHPArr[m].lodMdlCount == 1 then(msh.name = (MeshName) + " Layer 3") else (msh.name = (MeshName + " Layer 3" + " (LOD: " + l as string + ")"))
			msh.material = PolyMat

			max modify mode
			select msh
			
			addmodifier msh (Edit_Normals ()) ui:off
			msh.Edit_Normals.MakeExplicit selection:#{1..Normal_array.count}
			EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
			EN_setNormal = msh.Edit_Normals.SetNormal
			normID = #{}
			for v = 1 to Normal_array.count do
			(
				free normID
				EN_convertVS #{v} &normID
				for id in normID do EN_setNormal id Normal_array[v]
			)

			if BoneCount > 0 do(
					if FSHPArr[m].matrFlag == 0 then(
						$.transform = BoneFixArray[FSHPArr[m].fsklIndx]
						if RigModel != 2 do(
						skinMod = skin ()
						addModifier msh skinMod
						maxbone = getnodebyname BoneArray[FSHPArr[m].fsklIndx].name
						skinOps.addBone skinMod maxbone 0
						)
					) else ( if RigModel != 2 do(
					skinMod = skin ()
					boneIDMap = #()
					addModifier msh skinMod
					msh.Skin.weightAllVertices = false

					if FSHPArr[m].matrFlag == 1 do(
						for b in BoneArray do b.transform = matrix3 1
					)

					for i = 1 to BoneCount do
					(
						maxbone = getnodebyname BoneArray[i].name
						if i != BoneCount then (
							skinOps.addBone skinMod maxbone 0
						)else(
							skinOps.addBone skinMod maxbone 1
						)
					)
					
					local numSkinBones = skinOps.GetNumberBones skinMod
					for i = 1 to numSkinBones do
					(
						local boneName = skinOps.GetBoneName skinMod i 0
						for j = 1 to BoneCount do
						(
							if boneName == BoneArray[j].Name then
							(
								boneIDMap[j] = i
								j = BoneCount + 1
							)
						)
					) -- Thanks to sunnydavis for the fix!
					
					modPanel.setCurrentObject skinMod

				for i = 1 to Vert_array.count do(
					skinOps.SetVertexWeights skinMod i 1 1
					skinOps.unnormalizeVertex skinMod i true 
					skinOps.SetVertexWeights skinMod i 1 0
				)
					skinOps.RemoveZeroWeights skinMod
				for i = 1 to Vert_array.count do(
					skinOps.unnormalizeVertex skinMod i false
				)

					for i = 1 to Weight_array.count do (
						w = Weight_array[i]
						bi = #() --bone index array
						wv = #() --weight value array
						
						for j = 1 to w.boneids.count do
						(
							boneid = w.boneids[j]
							weight = w.weights[j]
							append bi boneIDMap[boneid]
							append wv weight
						)   
						skinOps.ReplaceVertexWeights skinMod i bi wv
					)

					if FSHPArr[m].matrFlag == 1 do(
						for b = 1 to BoneCount do BoneArray[b].transform = BoneFixArray[b]
					)

				)
				)
			)

		)
		
		if UV4_array.count > 1 do(
			local PolyMat = standardMaterial()
			PolyMat.name = (FMATNameArr[FSHPArr[m].fmatIndx] as string)
			PolyMat.showinviewport = true
			PolyMat.twosided = false
			fseek f (FMATArr[FSHPArr[m].fmatIndx].texSelOff)#seek_set
			TextureNameOffset = readoffset f
			fseek f TextureNameOffset#seek_set
			TextureName = readstring f
			if TexFormat == 2 then (tm = Bitmaptexture filename:(p + TextureName as string + ".png")) else (tm = Bitmaptexture filename:(p + TextureName as string + ".dds"))
			tm.alphasource = 0
			PolyMat.diffuseMap = tm
			PolyMat.opacityMap = tm
			PolyMat.opacityMap.monoOutput = 1
			
			msh = mesh vertices:Vert_array faces:Face_array
			msh.numTVerts = Vert_Array.count
			setNumCPVVerts msh msh.numTVerts
			setCVertMode msh true
			setShadeCVerts msh true
			defaultVCFaces msh
			buildTVFaces msh
			for j = 1 to Color_Array.count do setvertcolor msh j Color_array[j]
			If Vert_Array.count == UV4_Array.count Do (
				for j = 1 to Vert_array.count do setTVert msh j UV4_array[j]
				for j = 1 to Face_array.count do setTVFace msh j Face_array[j]
			)
			for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
				
			if FSHPArr[m].lodMdlCount == 1 then(msh.name = (MeshName) + " Layer 4") else (msh.name = (MeshName + " Layer 4" + " (LOD: " + l as string + ")"))
			msh.material = PolyMat

			select msh
			
			addmodifier msh (Edit_Normals ()) ui:off
			msh.Edit_Normals.MakeExplicit selection:#{1..Normal_array.count}
			EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
			EN_setNormal = msh.Edit_Normals.SetNormal
			normID = #{}
			for v = 1 to Normal_array.count do
			(
				free normID
				EN_convertVS #{v} &normID
				for id in normID do EN_setNormal id Normal_array[v]
			)

			if BoneCount > 0 do(
					if FSHPArr[m].matrFlag == 0 then(
						$.transform = BoneFixArray[FSHPArr[m].fsklIndx]
						if RigModel != 2 do(
						skinMod = skin ()
						addModifier msh skinMod
						maxbone = getnodebyname BoneArray[FSHPArr[m].fsklIndx].name
						skinOps.addBone skinMod maxbone 0
						)
					) else ( if RigModel != 2 do(
					skinMod = skin ()
					boneIDMap = #()
					addModifier msh skinMod
					msh.Skin.weightAllVertices = false

					if FSHPArr[m].matrFlag == 1 do(
						for b in BoneArray do b.transform = matrix3 1
					)

					for i = 1 to BoneCount do
					(
						maxbone = getnodebyname BoneArray[i].name
						if i != BoneCount then (
							skinOps.addBone skinMod maxbone 0
						)else(
							skinOps.addBone skinMod maxbone 1
						)
					)
					
					local numSkinBones = skinOps.GetNumberBones skinMod
					for i = 1 to numSkinBones do
					(
						local boneName = skinOps.GetBoneName skinMod i 0
						for j = 1 to BoneCount do
						(
							if boneName == BoneArray[j].Name then
							(
								boneIDMap[j] = i
								j = BoneCount + 1
							)
						)
					) -- Thanks to sunnydavis for the fix!
					
					modPanel.setCurrentObject skinMod

				for i = 1 to Vert_array.count do(
					skinOps.SetVertexWeights skinMod i 1 1
					skinOps.unnormalizeVertex skinMod i true 
					skinOps.SetVertexWeights skinMod i 1 0
				)
					skinOps.RemoveZeroWeights skinMod
				for i = 1 to Vert_array.count do(
					skinOps.unnormalizeVertex skinMod i false
				)

					for i = 1 to Weight_array.count do (
						w = Weight_array[i]
						bi = #() --bone index array
						wv = #() --weight value array
						
						for j = 1 to w.boneids.count do
						(
							boneid = w.boneids[j]
							weight = w.weights[j]
							append bi boneIDMap[boneid]
							append wv weight
						)   
						skinOps.ReplaceVertexWeights skinMod i bi wv
					)

					if FSHPArr[m].matrFlag == 1 do(
						for b = 1 to BoneCount do BoneArray[b].transform = BoneFixArray[b]
					)

				)
				)
			)

		)
		
		if UV5_array.count > 1 do(
			local PolyMat = standardMaterial()
			PolyMat.name = (FMATNameArr[FSHPArr[m].fmatIndx] as string)
			PolyMat.showinviewport = true
			PolyMat.twosided = false
			fseek f (FMATArr[FSHPArr[m].fmatIndx].texSelOff)#seek_set
			TextureNameOffset = readoffset f
			fseek f TextureNameOffset#seek_set
			TextureName = readstring f
			if TexFormat == 2 then (tm = Bitmaptexture filename:(p + TextureName as string + ".png")) else (tm = Bitmaptexture filename:(p + TextureName as string + ".dds"))
			tm.alphasource = 0
			PolyMat.diffuseMap = tm
			PolyMat.opacityMap = tm
			PolyMat.opacityMap.monoOutput = 1
			
			msh = mesh vertices:Vert_array faces:Face_array
			msh.numTVerts = Vert_Array.count
			setNumCPVVerts msh msh.numTVerts
			setCVertMode msh true
			setShadeCVerts msh true
			defaultVCFaces msh
			buildTVFaces msh
			for j = 1 to Color_Array.count do setvertcolor msh j Color_array[j]
			If Vert_Array.count == UV5_Array.count Do (
				for j = 1 to Vert_array.count do setTVert msh j UV5_array[j]
				for j = 1 to Face_array.count do setTVFace msh j Face_array[j]
			)
			for j = 1 to msh.numfaces do setFaceSmoothGroup msh j 1
				
			if FSHPArr[m].lodMdlCount == 1 then(msh.name = (MeshName) + " Layer 5") else (msh.name = (MeshName + " Layer 5" + " (LOD: " + l as string + ")"))
			msh.material = PolyMat

			select msh
			
			addmodifier msh (Edit_Normals ()) ui:off
			msh.Edit_Normals.MakeExplicit selection:#{1..Normal_array.count}
			EN_convertVS = msh.Edit_Normals.ConvertVertexSelection
			EN_setNormal = msh.Edit_Normals.SetNormal
			normID = #{}
			for v = 1 to Normal_array.count do
			(
				free normID
				EN_convertVS #{v} &normID
				for id in normID do EN_setNormal id Normal_array[v]
			)

			if BoneCount > 0 do(
					if FSHPArr[m].matrFlag == 0 then(
						$.transform = BoneFixArray[FSHPArr[m].fsklIndx]
						if RigModel != 2 do(
						skinMod = skin ()
						addModifier msh skinMod
						maxbone = getnodebyname BoneArray[FSHPArr[m].fsklIndx].name
						skinOps.addBone skinMod maxbone 0
						)
					) else ( if RigModel != 2 do(
					skinMod = skin ()
					boneIDMap = #()
					addModifier msh skinMod
					msh.Skin.weightAllVertices = false

					if FSHPArr[m].matrFlag == 1 do(
						for b in BoneArray do b.transform = matrix3 1
					)

					for i = 1 to BoneCount do
					(
						maxbone = getnodebyname BoneArray[i].name
						if i != BoneCount then (
							skinOps.addBone skinMod maxbone 0
						)else(
							skinOps.addBone skinMod maxbone 1
						)
					)
					
					local numSkinBones = skinOps.GetNumberBones skinMod
					for i = 1 to numSkinBones do
					(
						local boneName = skinOps.GetBoneName skinMod i 0
						for j = 1 to BoneCount do
						(
							if boneName == BoneArray[j].Name then
							(
								boneIDMap[j] = i
								j = BoneCount + 1
							)
						)
					) -- Thanks to sunnydavis for the fix!
					
					modPanel.setCurrentObject skinMod

				for i = 1 to Vert_array.count do(
					skinOps.SetVertexWeights skinMod i 1 1
					skinOps.unnormalizeVertex skinMod i true 
					skinOps.SetVertexWeights skinMod i 1 0
				)
					skinOps.RemoveZeroWeights skinMod
				for i = 1 to Vert_array.count do(
					skinOps.unnormalizeVertex skinMod i false
				)

					for i = 1 to Weight_array.count do (
						w = Weight_array[i]
						bi = #() --bone index array
						wv = #() --weight value array
						
						for j = 1 to w.boneids.count do
						(
							boneid = w.boneids[j]
							weight = w.weights[j]
							append bi boneIDMap[boneid]
							append wv weight
						)   
						skinOps.ReplaceVertexWeights skinMod i bi wv
					)

					if FSHPArr[m].matrFlag == 1 do(
						for b = 1 to BoneCount do BoneArray[b].transform = BoneFixArray[b]
					)

				)
				)
			)

		)
		)
		
		if UVLayers == 2 do(
			if UV2_array.count > 0 do(
			meshop.setNumMaps msh 3 keep:true
			for i = 1 to UV2_array.count do (
					meshop.setMapVert msh 2 i UV2_array[i]
				)
			)
			if UV3_array.count > 0 do(
			meshop.setNumMaps msh 4 keep:true
			for i = 1 to UV3_array.count do (
					meshop.setMapVert msh 3 i UV3_array[i]
				)
			)
			if UV4_array.count > 0 do(
			meshop.setNumMaps msh 5 keep:true
			for i = 1 to UV4_array.count do (
					meshop.setMapVert msh 4 i UV4_array[i]
				)
			)
			if UV5_array.count > 0 do(
			meshop.setNumMaps msh 6 keep:true
			for i = 1 to UV5_array.count do (
					meshop.setMapVert msh 5 i UV5_array[i]
				)
			)
		)

		fseek f LODRet #seek_set

		)
	)

	clearSelection()

Print ("Done! ("+((((timestamp())-st)*0.001)as string)+" Seconds)")
enableSceneRedraw()

)
)
)

CreateDialog BFRESImporter